/*********************************************************************/
/* file: files.c - funtions for logfile and reading/writing comfiles */
/*                             TINTIN + +                            */
/*          (T)he K(I)cki(N) (T)ickin D(I)kumud Clie(N)t             */
/*                     coded by peter unold 1992                     */
/*                    New code by Bill Reiss 1993                    */
/*********************************************************************/
#include "stdafx.h"

struct listnode *common_aliases, *common_actions, *common_subs, *common_myvars,
*common_highs, *common_antisubs, *common_pathdirs;
struct completenode *complete_head;

/**********************************/
/* load a completion file         */
/**********************************/
int read_complete(char *filename)
{
  FILE *myfile;
  char buffer[BUFFER_SIZE], *cptr;
  int flag;
  struct completenode *tcomplete, *tcomp2;

  flag = TRUE;

  if ((complete_head = (struct completenode *) (mymalloc(sizeof(struct completenode)))) == NULL)
  {
    syserr("couldn't alloc completehead");
  }

  tcomplete = complete_head;
	tcomplete->next = NULL;

  if ((myfile = fopen("tab.txt", "r")) == NULL)
		return 0;

  while (fgets(buffer, sizeof(buffer), myfile))
  {
    for (cptr = buffer; *cptr && *cptr != '\n'; cptr++) ;
    *cptr = '\0';

    if ((tcomp2 = (struct completenode *) (mymalloc(sizeof(struct completenode)))) == NULL)
    {
      syserr("couldn't alloc completenode");
    }

    if ((cptr = (char *) (mymalloc(strlen(buffer) + 1))) == NULL)
    {
      syserr("couldn't alloc memory for string in complete");
    }

    strcpy(cptr, buffer);
    tcomp2->strng = cptr;
    tcomplete->next = tcomp2;
    tcomplete = tcomp2;
  }

  tcomplete->next = NULL;
  fclose(myfile);
  sprintf(buffer, "tab completion file '%s' loaded.\n", filename);
  tintin_puts(buffer, (struct session *) NULL);
  prompt(NULL);
  tintin_puts("\n", (struct session *) NULL);
	return 1;
}

/********************/
/* the #log command */
/********************/
void log_command(char *arg, struct session *ses)
{
  if (ses)
  {
    if (!ses->logfile)
    {
      if (*arg)
      {
	if ((ses->logfile = fopen(arg, "w")))
	  tintin_puts("#OK. LOGGING.....", ses);
	else
	  tintin_puts("#COULDN'T OPEN FILE.", ses);
      }
      else
	tintin_puts("#SPECIFY A FILENAME.", ses);
    }
    else
    {
      fclose(ses->logfile);
      ses->logfile = NULL;
      tintin_puts("#OK. LOGGING TURNED OFF.", ses);
    }
  }
  else
    tintin_puts("#THERE'S NO SESSION TO LOG.", ses);

  prompt(NULL);
}

/***********************/
/* the #append command */
/***********************/
void append_command(char *arg, struct session *ses)
{
	FILE *logfile;
  char left[BUFFER_SIZE], right[BUFFER_SIZE], result[BUFFER_SIZE];

	arg = get_arg_in_braces(arg, left, 0);
  arg = get_arg_in_braces(arg, right, 1);

	if(!ses)
	{
		tintin_puts("#THERE'S NO SESSION.", ses);
		return;
	}

  if (!*left)
	{
		tintin_puts("#SPECIFY A FILENAME.", ses);
		return;
	}

  if (!*right)
	{
		tintin_puts("#NOTHING TO LOG.", ses);
		return;
	}

	if(!(logfile = fopen(left, "a")))
	{
		tintin_puts("#COULDN'T OPEN FILE.", ses);
		return;
	}
  
  prepare_actionalias(right, result, ses);

	fputs(result, logfile);
	fputs("\n", logfile);
	fclose(logfile);
  prompt(NULL);
}

/***********************************/
/* read and execute a command file */
/***********************************/
struct session *read_command(char *filename, struct session *ses)
{
  FILE *myfile;
  char buffer[BUFFER_SIZE], *cptr;
  char message[80];
  int flag;

  flag = TRUE;
  get_arg_in_braces(filename, filename, 1);

  if ((myfile = fopen(filename, "r")) == NULL)
  {
    tintin_puts("#ERROR - COULDN'T OPEN THAT FILE.", ses);
    prompt(NULL);
    return ses;
  }

  if (!verbose)
    puts_echoing = FALSE;

  alnum = 0;
  acnum = 0;
  subnum = 0;
  varnum = 0;
  hinum = 0;
  antisubnum = 0;

  while (fgets(buffer, sizeof(buffer), myfile))
  {
    if (flag)
    {
      puts_echoing = TRUE;
      char_command(buffer, ses);

      if (!verbose)
	puts_echoing = FALSE;

      flag = FALSE;
    }

    for (cptr = buffer; *cptr && *cptr != '\n'; cptr++)
      ;

    *cptr = '\0';

    ses = parse_input(buffer, ses);
  }

  if (!verbose)
  {
    puts_echoing = TRUE;
    sprintf(message, "#OK. %d ALIASES LOADED.", alnum);
    tintin_puts2(message, ses);
    sprintf(message, "#OK. %d ACTIONS LOADED.", acnum);
    tintin_puts2(message, ses);
    sprintf(message, "#OK. %d ANTISUBS LOADED.", antisubnum);
    tintin_puts2(message, ses);
    sprintf(message, "#OK. %d SUBSTITUTES LOADED.", subnum);
    tintin_puts2(message, ses);
    sprintf(message, "#OK. %d VARIABLES LOADED.", varnum);
    tintin_puts2(message, ses);
    sprintf(message, "#OK. %d HIGHLIGHTS LOADED.", hinum);
    tintin_puts2(message, ses);
  }

  fclose(myfile);
  prompt(NULL);

  return ses;
}

/*************************************************/
/* read and execute a SRC formatted command file */
/* Added by Greg Milford  4/2000                 */
/*************************************************/
struct session *readsrc_command(char *filename, struct session *ses)
{
  FILE *myfile;
  char buffer[BUFFER_SIZE], tempbuffer[BUFFER_SIZE], *cptr;
  char message[80];
  int flag;
  int num_opening = 0;
  int num_closing = 0;
  int line_number = 0;
  int bufferlength;
  int startlastcommand = 1;
  int i;
  
  //  tempbuffer needs this initial condition for the first command.
  tempbuffer[0] = '\0';

  flag = TRUE;
  get_arg_in_braces(filename, filename, 1);

  if ((myfile = fopen(filename, "r")) == NULL)
  {
    tintin_puts("#ERROR - COULDN'T OPEN THAT FILE.", ses);
    prompt(NULL);
    return ses;
  }

  if (!verbose)
    puts_echoing = FALSE;

  alnum = 0;
  acnum = 0;
  subnum = 0;
  varnum = 0;
  hinum = 0;
  antisubnum = 0;

  while (fgets(buffer, sizeof(buffer), myfile))
  {
	line_number++;

    if (flag)
    {
      puts_echoing = TRUE;
      char_command(buffer, ses);

      if (!verbose)
	puts_echoing = FALSE;

      flag = FALSE;
    }

	//  This for loop traverses a single line of the command file.
    for (cptr = buffer; *cptr && *cptr != '\n'; cptr++) {
      if (*cptr == '{')
        num_opening++;
	  if (*cptr == '}')
	    num_closing++;
	}

	// trim off spaces from ends of strings
	
	while (*(cptr - 1) == ' ') {
	  cptr--;
	}
    *cptr = '\0';

	// trim spaces of front of new string buffer
	cptr = buffer;
	
	while (*cptr == ' ') {
	  cptr++;
	}

	// the following lines talley up how long a command is getting
	// to catch a brackets error.

	bufferlength = 0;
    while (tempbuffer[bufferlength] != '\0') {
	  bufferlength++;
	}
	for (i = 0; buffer[i] != '\0';i++) {
	  bufferlength++;
	}

	//  This catches a buffer overflow before it happens and makes
	//  a good guess as to where the problem started.
	//  If it catches it, it will close the file and exit cleanly 
	//  back to Wintin.  All commands prior to the erroneous one
	//  were successfully read into the interpreter.
	
	if (bufferlength > 2040) {
	  puts_echoing = TRUE;
      sprintf(message, "#OK. %d ALIASES LOADED.", alnum);
      tintin_puts2(message, ses);
      sprintf(message, "#OK. %d ACTIONS LOADED.", acnum);
      tintin_puts2(message, ses);
      sprintf(message, "#OK. %d ANTISUBS LOADED.", antisubnum);
      tintin_puts2(message, ses);
      sprintf(message, "#OK. %d SUBSTITUTES LOADED.", subnum);
      tintin_puts2(message, ses);
      sprintf(message, "#OK. %d VARIABLES LOADED.", varnum);
      tintin_puts2(message, ses);
      sprintf(message, "#OK. %d HIGHLIGHTS LOADED.", hinum);
      tintin_puts2(message, ses);
	  sprintf(message, " ");
      tintin_puts2(message, ses);
	  sprintf(message, "Buffer overflow on line %d of the command file.", line_number);
	  tintin_puts2(message, ses);
  	  sprintf(message, "Check for a Brackets error possibly in your Wintin command starting at line %d of your command file.", startlastcommand);
	  tintin_puts2(message, ses);
      fclose(myfile);
	  prompt(NULL);
	  return ses;
	}
   

	  //  Concatenate buffer to what previously existed
	strcat(tempbuffer, cptr);
	  //  make sure cptr is pointing to end of full buffer when exiting.
	if ((num_opening == num_closing) && (num_closing > 1)) {
	  num_closing = 0;
	  num_opening = 0;
	  ses = parse_input(tempbuffer, ses);
	  tempbuffer[0] = '\0';
	  startlastcommand = line_number + 1;
	}
  }

  if (!verbose)
  {
    puts_echoing = TRUE;
    sprintf(message, "#OK. %d ALIASES LOADED.", alnum);
    tintin_puts2(message, ses);
    sprintf(message, "#OK. %d ACTIONS LOADED.", acnum);
    tintin_puts2(message, ses);
    sprintf(message, "#OK. %d ANTISUBS LOADED.", antisubnum);
    tintin_puts2(message, ses);
    sprintf(message, "#OK. %d SUBSTITUTES LOADED.", subnum);
    tintin_puts2(message, ses);
    sprintf(message, "#OK. %d VARIABLES LOADED.", varnum);
    tintin_puts2(message, ses);
    sprintf(message, "#OK. %d HIGHLIGHTS LOADED.", hinum);
    tintin_puts2(message, ses);
  }

  fclose(myfile);
  prompt(NULL);

  return ses;
}



/************************/
/* write a command file */
/************************/
struct session *write_command(char *filename, struct session *ses)
{
  FILE *myfile;
  char buffer[BUFFER_SIZE];
  struct listnode *nodeptr;

  get_arg_in_braces(filename, filename, 1);

  if (*filename == '\0')
  {
    tintin_puts("#ERROR - COULDN'T OPEN THAT FILE.", ses);
    prompt(NULL);
    return (0);			/* added zero return */
  }

  if ((myfile = fopen(filename, "w")) == NULL)
  {
    tintin_puts("#ERROR - COULDN'T OPEN THAT FILE.", ses);
    prompt(NULL);
    return (0);			/* added zero return */
  }

  nodeptr = (ses) ? ses->aliases : common_aliases;

  while ((nodeptr = nodeptr->next))
  {
    prepare_for_write("alias", nodeptr->left, nodeptr->right, "\0", buffer);
    fputs(buffer, myfile);
  }

  nodeptr = (ses) ? ses->actions : common_actions;

  while ((nodeptr = nodeptr->next))
  {
    prepare_for_write("action", nodeptr->left, nodeptr->right, nodeptr->pr, buffer);
    fputs(buffer, myfile);
  }

  nodeptr = (ses) ? ses->antisubs : common_antisubs;

  while ((nodeptr = nodeptr->next))
  {
    prepare_for_write("antisubstitute", nodeptr->left,
		      nodeptr->right, "\0", buffer);
    fputs(buffer, myfile);
  }

  nodeptr = (ses) ? ses->subs : common_subs;

  while ((nodeptr = nodeptr->next))
  {
    prepare_for_write("substitute", nodeptr->left, nodeptr->right, "\0", buffer);
    fputs(buffer, myfile);
  }

  nodeptr = (ses) ? ses->myvars : common_myvars;

  while ((nodeptr = nodeptr->next))
  {
    prepare_for_write("variable", nodeptr->left, nodeptr->right, "\0", buffer);
    fputs(buffer, myfile);
  }

  nodeptr = (ses) ? ses->highs : common_highs;

  while ((nodeptr = nodeptr->next))
  {
    prepare_for_write("highlight", nodeptr->right, nodeptr->left, "\0", buffer);
    fputs(buffer, myfile);
  }

  nodeptr = (ses) ? ses->pathdirs : common_pathdirs;

  while ((nodeptr = nodeptr->next))
  {
    prepare_for_write("pathdir", nodeptr->right, nodeptr->left, "\0", buffer);
    fputs(buffer, myfile);
  }

  fclose(myfile);
  tintin_puts("#COMMAND-FILE WRITTEN.", ses);

  return ses;
}

/************************/
/* write a command file */
/************************/
struct session *writesession_command(char *filename, struct session *ses)
{
  FILE *myfile;
  char buffer[BUFFER_SIZE];
  struct listnode *nodeptr;

  get_arg_in_braces(filename, filename, 1);

  if (*filename == '\0')
  {
    tintin_puts("#ERROR - COULDN'T OPEN THAT FILE.", ses);
    prompt(NULL);
    return (0);			/* added zero return */
  }
  if ((myfile = fopen(filename, "w")) == NULL)
  {
    tintin_puts("#ERROR - COULDN'T OPEN THAT FILE.", ses);
    prompt(NULL);
    return (0);			/* added zero return */
  }

  nodeptr = (ses) ? ses->aliases : common_aliases;

  while ((nodeptr = nodeptr->next))
  {
    if (ses && searchnode_list(common_aliases, nodeptr->left))
      continue;

    prepare_for_write("alias", nodeptr->left, nodeptr->right, "\0", buffer);
    fputs(buffer, myfile);
  }

  nodeptr = (ses) ? ses->actions : common_actions;

  while ((nodeptr = nodeptr->next))
  {
    if (ses && searchnode_list(common_actions, nodeptr->left))
      continue;

    prepare_for_write("action", nodeptr->left, nodeptr->right, nodeptr->pr, buffer);
    fputs(buffer, myfile);
  }

  nodeptr = (ses) ? ses->antisubs : common_antisubs;

  while ((nodeptr = nodeptr->next))
  {
    if (ses && searchnode_list(common_antisubs, nodeptr->left))
      continue;

    prepare_for_write("antisubstitute", nodeptr->left, "", "\0", buffer);
    fputs(buffer, myfile);
  }

  nodeptr = (ses) ? ses->subs : common_subs;

  while ((nodeptr = nodeptr->next))
  {
    if (ses && searchnode_list(common_subs, nodeptr->left))
      continue;

    prepare_for_write("substitute", nodeptr->left, nodeptr->right, "\0", buffer);
    fputs(buffer, myfile);
  }

  nodeptr = (ses) ? ses->myvars : common_myvars;

  while ((nodeptr = nodeptr->next))
  {
    if (ses && searchnode_list(common_myvars, nodeptr->left))
      continue;

    prepare_for_write("variable", nodeptr->left, nodeptr->right, "\0", buffer);
    fputs(buffer, myfile);
  }

  nodeptr = (ses) ? ses->highs : common_highs;

  while ((nodeptr = nodeptr->next))
  {
    if (ses && searchnode_list(common_highs, nodeptr->left))
      continue;

    prepare_for_write("highlight", nodeptr->right, nodeptr->left, "\0", buffer);
    fputs(buffer, myfile);
  }

  fclose(myfile);
  tintin_puts("#COMMANDO-FILE WRITTEN.", ses);
  return ses;
}

void prepare_for_write(char *command, char *left, char *right, char *pr, char *result)
{
  *result = tintin_char;
  *(result + 1) = '\0';
  strcat(result, command);
  strcat(result, " {");
  strcat(result, left);
  strcat(result, "}");

  if (strlen(right) != 0)
  {
    strcat(result, " {");
    strcat(result, right);
    strcat(result, "}");
  }

  if (strlen(pr) != 0)
  {
    strcat(result, " {");
    strcat(result, pr);
    strcat(result, "}");
  }

  strcat(result, "\n");
}

void prepare_quotes(char *string)
{
  char s[BUFFER_SIZE], *cpsource, *cpdest;
  int nest = FALSE;

  strcpy(s, string);

  cpsource = s;
  cpdest = string;

  while (*cpsource)
  {
    if (*cpsource == '\\')
    {
      *cpdest++ = *cpsource++;

      if (*cpsource)
	*cpdest++ = *cpsource++;
    }
    else if (*cpsource == '\"' && nest == FALSE)
    {
      *cpdest++ = '\\';
      *cpdest++ = *cpsource++;
    }
    else if (*cpsource == '{')
    {
      nest = TRUE;
      *cpdest++ = *cpsource++;
    }
    else if (*cpsource == '}')
    {
      nest = FALSE;
      *cpdest++ = *cpsource++;
    }
    else
      *cpdest++ = *cpsource++;
  }

  *cpdest = '\0';
}
