/*
 * Screen buffer class definition file
 * 
 * WL 28/11/94
 */  
                   
#include "stdafx.h"                   
#include "scrbuf.h"
#include "enhstr.h"
#include "colorstate.h"

extern int CrackAnsi(const char *str, int *format, int *constituents);



/*
 * Constructor/Destructor
 */

CScreenBuffer::CScreenBuffer()
{
	LineBuffer = new (CString( *[DEFAULT_BUFFER_SIZE]));
	MaxLines = DEFAULT_BUFFER_SIZE;
	LastLine = Current = 0;
	InsertLine = InsertPos = 0;
	InsertMode = 1;  
	AddLine("");
}

CScreenBuffer::CScreenBuffer(int iSize)
{
	LineBuffer = new (CString( *[iSize]));
	MaxLines = iSize;
	LastLine = Current = 0;
	InsertLine = InsertPos = 0;
	InsertMode = 1;  
	AddLine("");
}

CScreenBuffer::~CScreenBuffer()
{
	unsigned i;

	for(i=0; i<LastLine; i++)
		FreeLineBuffer(LineBuffer[i]);

	delete [] LineBuffer;
}
 
 
/*
 * Public Functions
 */


// Return the current line as a CString

CString *CScreenBuffer::GetCurrent(void)
{
/*	CEnhString cs;
	cs.printf("Current %d LastLine %d\n", Current, LastLine);
	TRACE(cs);
*/	
	if(Current >= LastLine)
	{
		char buf[90];
		sprintf(buf, "Current invalid: Current %d LastLine %d\n", Current, LastLine);
		TRACE(buf);

		if(LastLine == 0)
			Current = 0;
		else
			Current = LastLine - 1;
//		return FALSE;	// buffer is empty
	}

	return LineBuffer[Current];
}

/*
 * Return the current line position
 */

long CScreenBuffer::GetCurrentLine(void)
{
	return Current;
}

/*
 * Return the insert line position
 */

long CScreenBuffer::GetInsertLine(void)
{
	return InsertLine;
}


// Set the current line as an offset from start, end or current

BOOL CScreenBuffer::SetCurrentLine(int iType, long lOffset)
{
	if(!LastLine)
		return FALSE;	// buffer is empty

	long c;

	switch(iType)
	{
	case SBSTART:
					c = lOffset;
					break;
	case SBEND:
					c = LastLine -1 + lOffset;
					break;
	case SBCURR:
					c = Current + lOffset;
					break;
	default:
					break;
	}

	if(c < 0)
	{
		Current = 0;
		return TRUE;	// just haven't got enough buffer
	}
	else if(c >= (long)LastLine)
	{
		Current = LastLine - 1;
		return FALSE;
	}
	else
	{
		Current = (unsigned) c;
		return TRUE;
	}
}

// Return the number of lines currently buffered

long CScreenBuffer::Range(void)
{
	return LastLine;
}

// Format a color string from colorstate struct

/*
void CScreenBuffer::MakeColorString(struct colorstate *cs, char *buf)
{
	char buf1[20];
	strcpy(buf, "\x1b[");

	sprintf(buf1, "%d;", cs->Fore);
	strcat(buf, buf1);
	sprintf(buf1, "%d;", cs->Back);
	strcat(buf, buf1);

	if(cs->Bold)
		strcat(buf, "1;");

	if(cs->Light)
		strcat(buf, "2;");

	if(strlen(buf) > 2)
	{
		buf[strlen(buf)-1] = 0;
		strcat(buf, "m");
	}
	else
		buf[0] = 0;
}
*/

// Add a line to the end of the buffer

BOOL CScreenBuffer::AddLine(const char *pcBuf)
{
	CString csBuf;
	ColorState cs;

	if(LastLine == MaxLines)
		DisposeLines();

	if(Current > LastLine)
		return FALSE;		// buffer empty

	// Sort out colour scheme

	if(LastLine > 0)
		cs.FindLast(LineBuffer[LastLine-1]);
	else
		cs.Reset = 1;

	cs.FormatAnsi(&csBuf);

	if(csBuf[0])
	{
		if((LineBuffer[LastLine] = GetLineBuffer(csBuf)) == NULL)
			return FALSE;

		*LineBuffer[LastLine] += pcBuf;
	}
	else
	{
		if((LineBuffer[LastLine] = GetLineBuffer(pcBuf)) == NULL)
			return FALSE;
	}

	InsertPos = LineBuffer[LastLine]->GetLength();
	LastLine++;

	Current = LastLine - 1;

	return TRUE;
}

// Replace the current line

BOOL CScreenBuffer::ReplaceCurrent(char *pcBuf)
{
	CString csBuf;
	ColorState cs;
	
	if(Current > LastLine)
		return FALSE;		// buffer empty

	// Sort out colour scheme

	cs.FindLast(LineBuffer[Current]);
	cs.FormatAnsi(&csBuf);

	FreeLineBuffer(LineBuffer[Current]);

	if(csBuf[0])
	{
		if((LineBuffer[Current] = GetLineBuffer(csBuf)) == NULL)
			return FALSE;

		*LineBuffer[Current] += pcBuf;
	}
	else
	{
		if((LineBuffer[Current] = GetLineBuffer(pcBuf)) == NULL)
			return FALSE;
	}

	InsertPos = LineBuffer[Current]->GetLength();
	return TRUE;
}

/*
 * Return the current insertion point in the line
 */
 
int CScreenBuffer::GetInsertPos()
{
	if(!LastLine)
		return 0;		// buffer is empty
		
	return InsertPos;	
} 

/*
 * Set the current insertion point in the line
 */

BOOL CScreenBuffer::SetInsertPos(int iType, int iOffset)
{
	if(!LastLine)
		return FALSE;	// buffer is empty

	switch(iType)
	{
	case SBSTART:
					InsertPos = iOffset;
					break;
	case SBEND:
					InsertPos = LineBuffer[InsertLine]->GetLength() -1 + iOffset;
					break;
	case SBCURR:
					InsertPos += iOffset;
					break;
	default:
					break;
	}

	if(InsertPos < 0)
	{
		InsertPos = 0;
		return FALSE;
	}
	else if(InsertPos > (unsigned)LineBuffer[InsertLine]->GetLength())
	{
		InsertPos = LineBuffer[InsertLine]->GetLength() - 1;
		return FALSE;
	}

	return TRUE;
	
}

/* 
 * Insert a character at the current insertion point, 
 * advance the insertion point
 */

BOOL CScreenBuffer::InsertChar(char c)
{
	unsigned n;
	unsigned iCount;
	int i, j;
	char d, buf[40];

	if(InsertLine > LastLine)
		return FALSE;

	if(c == 0x0d)		// CR
	{
		InsertPos = 0;

		// Skip over an initial ansi sequence, if present

		if((iCount = LineBuffer[InsertLine]->GetLength()) > 3)
		{
			
			InsertPos = 0;
			unsigned i = 0;
			 
			while(i < iCount-3)
			{ 
				if(LineBuffer[InsertLine]->GetAt(i) != 0x1b
				   || LineBuffer[InsertLine]->GetAt(i+1) != '[')
				   	return TRUE;

				for(i = i + 2; i < iCount; i++)
				{
					j = LineBuffer[InsertLine]->GetAt(i);
					if(!((j >= '0' && j <= '9') || j == ';'))
						break;
				}

				if(i  < iCount && j == 'm')
					InsertPos = ++i;
				else
					break;
			}
		}

		return TRUE;
	}
	else if(c == 0x0a)	// LF
	{
		AddLine("");
		InsertLine = LastLine - 1;
		return TRUE;
	}
	else if(c == 0x08)  // Backspace
	{
		if(InsertPos > 0)
		{
			if(InsertPos < ( n = LineBuffer[InsertLine]->GetLength()))
			{
				*LineBuffer[InsertLine] = LineBuffer[InsertLine]->Left(InsertPos-1) 
					 + LineBuffer[InsertLine]->Right(n-InsertPos);
			}
			else
			{
				*LineBuffer[InsertLine] = LineBuffer[InsertLine]->Left(InsertPos-1);
			}

			InsertPos--;
		}
		return TRUE;
	}
	else if(c == 'm')
	{
		// this is ugly. There's a possibility that this 'm' terminates
		// an ANSI escape sequence, in which case we need to fish it out and
		// feed it to InsertAnsi. We also need to check whether or not there
		// IS an ANSI escape sequence here without spinning too many cycles.

		if(InsertPos > 3)
		{
			d = LineBuffer[InsertLine]->GetAt(InsertPos-1);
			
			if(d >= '0' && d <= '9')
			{
				d = LineBuffer[InsertLine]->GetAt(InsertPos-2);

				if((d >= '0' && d <= '9') || d == ';' || d == '[')
				{
					// ok there's a fair chance there's an ANSI sequence here
					// let's burn some serious cycles

					i = LineBuffer[InsertLine]->ReverseFind(0x1b);

					if(i >= 0 && i < (int)InsertPos && LineBuffer[InsertLine]->GetAt(i+1) == '[')
					{
						for(j = InsertPos-1; j > i+1; j--)
						{
							d = LineBuffer[InsertLine]->GetAt(j);
							
							if(!((d >= '0' && d <= '9') || d == ';'))
								break;
						}

						if(j == i+1)
						{
							// it's an escape sequence alright. Chop it out and feed
							// it to InsertAnsi.


							strncpy(buf, LineBuffer[InsertLine]->Mid(i, InsertPos-i), 38);
							strcat(buf, "m");
							
							if(InsertPos < ( n = LineBuffer[InsertLine]->GetLength()))
							{
								// Insert

									*LineBuffer[InsertLine] = LineBuffer[InsertLine]->Left(i)  
														 + LineBuffer[InsertLine]->Right(n-InsertPos);
							}
							else
							{
								*LineBuffer[InsertLine] = LineBuffer[InsertLine]->Left(i);
							}

							InsertPos = i;

							InsertAnsi(buf);
							return TRUE;
						}
					}
				}
			}
		}
	}

	if(InsertPos < ( n = LineBuffer[InsertLine]->GetLength()))
	{
		// Insert

		if(InsertMode)
		{
			*LineBuffer[InsertLine] = LineBuffer[InsertLine]->Left(InsertPos) + c 
								 + LineBuffer[InsertLine]->Right(n-InsertPos);
		}
		else
		{
			LineBuffer[InsertLine]->SetAt(InsertPos, c);
		}

		InsertPos++;
	}
	else
	{
		*LineBuffer[InsertLine] += c;
		InsertPos++;
	}

	return TRUE;
}

/*
 * Insert an ANSI sequence at the current insertion point.
 *
 * All ANSI sequences should be inserted this way rather than
 * InsertChar. This function will 'fix up' the ANSI sequence
 * so that it contains all details of the current colour
 * scheme (foreground, background etc) rather than just
 * the changes in the control sequence it is passed. This
 * full information makes it possible for the viewer to work
 * without retaining state.
 *
 * This function is called automatically by InsertChar,
 * provided all additions to the screen buffer are made
 * via InsertChar you do not need to call this function
 * directly.
 */

BOOL CScreenBuffer::InsertAnsi(const char *pcBuf)
{
	int iCount;
	unsigned n;
	CString csBuf;
	ColorState cs;
	
	iCount = cs.CrackAnsi(pcBuf);

	if(iCount <= 0)
		return 0;		// for what it's worth, we didn't hit a valid sequence

	if(iCount == 99)
	{
		// this is the special 'revert' code. Need to find the previous state

		cs.FindPenul(LineBuffer[InsertLine]);
	}
	else
	{
		cs.FindLast(LineBuffer[InsertLine]);
	}

	cs.MergeAnsi(pcBuf);
	cs.FormatAnsi(&csBuf);

	// buf holds the sequence for us to insert

	if(InsertPos < ( n = LineBuffer[InsertLine]->GetLength()))
	{
		// Insert

		*LineBuffer[InsertLine] = LineBuffer[InsertLine]->Left(InsertPos) + csBuf 
								 + LineBuffer[InsertLine]->Right(n-InsertPos);
	}
	else
	{
		*LineBuffer[InsertLine] += csBuf;
	}

	InsertPos += csBuf.GetLength();
	return 1;
}


/*
 * Private Functions
 */

CString * CScreenBuffer::GetLineBuffer(const char *pcBuf)
{
	CString *pCS = new CString(pcBuf);

	while(LastLine && (pCS == NULL))
		DisposeLines();

	return pCS;	
}


// Free the specified line buffer

BOOL CScreenBuffer::FreeLineBuffer(CString *pCS)
{
	delete pCS;
	return TRUE;
}

// Dispose of a few lines from the start of the buffer to free up memory

BOOL CScreenBuffer::DisposeLines()
{
	unsigned i;
	unsigned long iLimit = MaxLines/10;

	/* Check for overrun (can happen if max is trimmed) */
	
	if(LastLine > MaxLines)
		iLimit += LastLine - MaxLines;

	if(iLimit >= LastLine)
		iLimit = LastLine - 1;

	for(i=0; i<iLimit; i++)
		FreeLineBuffer(LineBuffer[i]);

	LastLine -= iLimit;

	Current -= iLimit;

	if(Current < 0)
		Current = 0;

	InsertLine = LastLine - 1;

	if(InsertLine < 0)
		InsertLine = 0;

	for(i=0; i<LastLine; i++)
		LineBuffer[i] = LineBuffer[i + iLimit];
		
	return TRUE;
}


// A class member to process the effect of colour sequences. Takes
// the previous colour state and the ANSI sequence, and returns
// the new colour state. Ideally, colour state should be a class
// and this should be a member function...

/*
void CScreenBuffer::ParseAnsi(char *str, struct colorstate *csPrev, struct colorstate *csNext)
{
	int i, iCount;
	int format[5];
	CString csTemp;

	csNext->Fore = csPrev->Fore;
	csNext->Back = csPrev->Back;
	csNext->Bold = csPrev->Bold;
	csNext->Light = csPrev->Light;

	i = CrackAnsi(str, format, &iCount);
	
	if(i <= 0 || iCount <= 0)
	{
		// empty or corrupt control string

		return;
	}

	if(format[0] == 99)
	{
		// special 'revert to previous' code
		// should never get here - these should have been stripped
		// out by the input routine

		return;
	}
	else
	{
		for(i = 0; i < iCount; i++)
		{
			if(format[i] == 1)
			{
				csNext->Bold = 1;
			}
			else if(format[i] == 2)
			{
				csNext->Light = 1;
			}
			else if(format[i] >= 30 && format[i] <= 37)
			{
				csNext->Fore = format[i];
			}
			else if(format[i] >= 40 && format[i] <= 47)
			{
				csNext->Back = format[i];
			}
		}
	}
}
*/

// Set up a control sequence to reflect the starting colour state for 
// specified line, based on the previous line or default if none.

/*
void CScreenBuffer::GetColorsForLine(int line, struct colorstate *csNext)
{
	struct colorstate csPrev;

	csNext->Back = csPrev.Back = 40;
	csNext->Fore = csPrev.Fore = 37;
	csNext->Bold = csPrev.Bold = 0;
	csNext->Light = csPrev.Light = 0;

	if(line > 0)
	{		
		char *pc = LineBuffer[line-1]->GetBuffer(1);

		while((pc = strstr(pc, "\x1b[")) != NULL)
		{
			csPrev.Back = csNext->Back;
			csPrev.Fore = csNext->Fore;
			csPrev.Bold = csNext->Bold;
			csPrev.Light = csNext->Light;

			ParseAnsi(pc++, &csPrev, csNext);
		}

		LineBuffer[line-1]->ReleaseBuffer();
	}

	// sanity check

	if(csNext->Fore + 10 == csNext->Back)
	{
		csNext->Back = 40;
		csNext->Fore = 37;
	}
}

*/

// #highlight uses a special ANSI escape code to revert to the
// previous colour code for a line. This function calculates
// the previous code to use. This function is very similar
// to GetColorStringForLine but it seems crufty to combine them.

/*
void CScreenBuffer::GetPreviousStringForLine(int line, char *buf)
{
	struct colorstate csPrev, csNext;

	csNext.Back = csPrev.Back = 40;
	csNext.Fore = csPrev.Fore = 37;
	csNext.Bold = csPrev.Bold = 0;
	csNext.Light = csPrev.Light = 0;

	if(line >= 0)
	{		
		char *pc = LineBuffer[line]->GetBuffer(1);

		while((pc = strstr(pc, "\x1b[")) != NULL)
		{
			csPrev.Back = csNext.Back;
			csPrev.Fore = csNext.Fore;
			csPrev.Bold = csNext.Bold;
			csPrev.Light = csNext.Light;
			ParseAnsi(pc++, &csPrev, &csNext);
		}

		LineBuffer[line]->ReleaseBuffer();
	}

	// sanity check

	if(csPrev.Fore + 10 == csPrev.Back)
	{
		csPrev.Back = 40;
		csPrev.Fore = 37;
	}

	MakeColorString(&csPrev, buf);
}
*/
/* Set the maximum number of lines in the buffer */

void CScreenBuffer::SetMaxLines(unsigned iMax)
{
	unsigned iNewMax = min(iMax, DEFAULT_BUFFER_SIZE);
	
	if(iNewMax < MaxLines)
	{
		MaxLines = iNewMax;
		DisposeLines();
	}
	else
	{
		MaxLines = iNewMax;
	}
}
